import unittest

from ginsfsm.gconfig import (
    GConfig,
    GConfigTemplateError,
    GConfigValidateError,
)


class TestGConfig(unittest.TestCase):
    def setUp(self):
        self.gconfig = GConfig({})

    def test_set_bad_argument(self):
        self.assertRaises(GConfigTemplateError, GConfig, '')
        self.assertRaises(GConfigTemplateError, GConfig, {0: []})
        self.assertRaises(GConfigTemplateError, GConfig, {'': []})
        self.assertRaises(GConfigTemplateError, GConfig, {'x': []})
        self.assertRaises(GConfigTemplateError, GConfig, {'x': 0})
        self.assertRaises(GConfigTemplateError, GConfig, {'x': [0]})
        self.assertRaises(GConfigTemplateError, GConfig,
                          {'x': [0, 0, 0, 0, '']})
        self.assertRaises(GConfigTemplateError, GConfig,
                          {'x': [GConfig, 0, 0, None, '']})

        def p():
            pass
        self.assertRaises(GConfigTemplateError, GConfig, {'x': [0, 0, 0, p, 0]})

        conf = GConfig({'x': [int, 0, 0, None, '']})
        self.assertTrue(isinstance(conf, GConfig))
        self.assertTrue(hasattr(conf.config, 'x'))
        value = conf.read_parameter('x')
        self.assertEqual(value, 0)

        conf = GConfig({'y': [int, -2, 0, None, '']})
        self.assertTrue(isinstance(conf, GConfig))
        self.assertTrue(hasattr(conf.config, 'y'))
        value = conf.read_parameter('y')
        self.assertEqual(value, -2)

        conf = GConfig(
            {
                'bt': [bool, "True", 0, None, ''],
                'bf': [bool, "False", 0, None, ''],
                'bn': [bool, None, 0, None, ''],
                'x': [int, 100, 0, None, ''],
                'y': [int, -2, 0, None, ''],
                's': [str, "kk", 0, None, ''],
                'None': [str, None, 0, None, ''],
                'ss': [list, "kk,mm", 0, None, ''],
                'pp': [tuple, "kk,mm", 0, None, ''],
            }
        )
        self.assertTrue(isinstance(conf, GConfig))
        value = conf.read_parameter('x')
        self.assertEqual(value, 100)
        value = conf.read_parameter('y')
        self.assertEqual(value, -2)
        value = conf.read_parameter('s')
        self.assertEqual(value, 'kk')
        value = conf.read_parameter('bt')
        self.assertEqual(value, True)
        value = conf.read_parameter('bf')
        self.assertEqual(value, False)
        value = conf.read_parameter('bn')
        self.assertEqual(value, False)
        value = conf.read_parameter('None')
        self.assertEqual(value, None)

        conf.write_parameters(bt=False)
        value = conf.read_parameter('bt')
        self.assertEqual(value, False)

        conf.reset_parameter('bt')
        value = conf.read_parameter('bt')
        self.assertEqual(value, True)

        value = conf.read_parameter('ss')
        self.assertEqual(value, ['kk', 'mm'])

        value = conf.read_parameter('pp')
        self.assertEqual(value, ('kk', 'mm'))

        conf = GConfig(
            {
                'lst': [list, ['', 0], 0, None, ''],
                'tup': [tuple, ('', 0), 0, None, ''],
            }
        )
        self.assertTrue(isinstance(conf, GConfig))
        value = conf.read_parameter('lst')
        self.assertEqual(value, ['', 0])
        value = conf.read_parameter('tup')
        self.assertEqual(value, ('', 0))
        #conf.write_parameters(xx=False)

        definition = [tuple, ('', 0), 0, None, '']
        conf = GConfig(
            {
                'tup': definition,
            }
        )
        conf.write_parameters(tup=(0, 0))
        definition[0] = GConfig
        conf.write_parameters(tup=(0, 0))

        conf = GConfig(
            {
                'lst': [list, ['', 0], 0, None, ''],
                'tup': [tuple, ('', 0), 0, None, ''],
            }
        )
        conf.read_parameters()

        conf = GConfig(
            {
                'x': [None, GConfig, 0, None, ''],
            }
        )
        value = conf.read_parameter('x')
        self.assertTrue(value is GConfig)

        def validate1(value):
            if value < -5:
                raise ValueError('Value too low %r' % value)
            if value < 0:
                return 0
            return value

        self.assertRaises(
            GConfigValidateError,
            GConfig,
            {
                'y': [int, -10, 0, validate1, '']
            }
        )
        value = conf.read_parameter('y')
        self.assertEqual(value, None)

        conf = GConfig({'y': [int, 10, 0, validate1, '']})
        value = conf.read_parameter('y')
        self.assertEqual(value, 10)
        conf.write_parameters(y=20)
        value = conf.read_parameter('y')
        self.assertTrue(value is 20)
        conf.reset_all_parameters()
        value = conf.read_parameter('y')
        self.assertTrue(value is 10)
        value = conf.read_parameter('yyy')
        self.assertTrue(value is None)

        def validate2(value):
            """
            sca1, sca+sca, http://localhost:8002+http://localhost:8002;
            sca2, sca+sca, http://localhost:8003+http://localhost:8003;

            must convert in:

            ((
                'sca1',
                ('sca','sca'),
                ('http://localhost:8002', 'http://localhost:8002'),
            ),
            (
                'sca2',
                ('sca','sca'),
                ('http://localhost:8003', 'http://localhost:8003'),
            ))

            """
            routes = []
            if value:
                value = value.split(';')
                for r in value:
                    r = r.strip()
                    if r:
                        routes.append(r)

                for idx, r in enumerate(routes):
                    items = r.split(',')
                    v = []
                    for it in items:
                        it = it.strip()
                        if it:
                            v.append(it)
                    routes[idx] = v

                for r in routes:
                    for idx, field in enumerate(r):
                        v = []
                        if not '+' in field:
                            v.append(field)
                            continue
                        items = field.split('+')
                        for e in items:
                            e = e.strip()
                            if e:
                                v.append(e)
                        r[idx] = tuple(v)

                for idx, r in enumerate(routes):
                    routes[idx] = tuple(r)

            return routes

        conf = GConfig(
            {
                'x': [tuple, None, 0, validate2, ''],
            }
        )
        s = """
            sca1, sca+sca, http://localhost:8002+http://localhost:8002;
            sca2, sca+sca, http://localhost:8003+http://localhost:8003;
            """
        d = (
            (
                'sca1',
                ('sca', 'sca'),
                ('http://localhost:8002', 'http://localhost:8002'),
            ),
            (
                'sca2',
                ('sca', 'sca'),
                ('http://localhost:8003', 'http://localhost:8003'),
            )
        )

        conf.write_parameters(x=s)
        x = conf.read_parameter('x')
        self.assertEqual(x, d)



import binascii
import os

from ginsfsm.compat import native_

from ginsfsm.scaffolds.template import Template  # API


class GinsFSMTemplate(Template):
    """
     A class that can be used as a base class for GinsFSM scaffolding
     templates.
    """
    def pre(self, command, output_dir, vars):
        """ Overrides :meth:`ginsfsm.scaffold.template.Template.pre`, adding
        several variables to the default variables list (including
        ``random_string``, and ``package_logger``).  It also prevents common
        misnamings (such as naming a package "site" or naming a package
        logger "root".
        """
        if vars['package'] == 'site':
            raise ValueError('Sorry, you may not name your package "site". '
                             'The package name "site" has a special meaning in '
                             'Python.  Please name it anything except "site".')
        vars['random_string'] = native_(binascii.hexlify(os.urandom(20)))
        package_logger = vars['package']
        if package_logger == 'root':
            # Rename the app logger in the rare case a project is named 'root'
            package_logger = 'app'
        vars['package_logger'] = package_logger
        return Template.pre(self, command, output_dir, vars)

    def post(self, command, output_dir, vars):  # pragma: no cover
        """ Overrides :meth:`ginsfsm.scaffold.template.Template.post`, to
        print "Welcome to GinsGSM.  Sorry for the convenience." after a
        successful scaffolding rendering."""
        self.out('Welcome to GinsFSM.  Sorry for the convenience.')
        return Template.post(self, command, output_dir, vars)

    def out(self, msg):  # pragma: no cover (replaceable testing hook)
        print(msg)


class MultipleWsgiTemplate(GinsFSMTemplate):
    _template_dir = 'multiple_wsgi'
    summary = 'Create a multiple WSGI proyect.'


class MultiplePyramidWsgiTemplate(GinsFSMTemplate):
    _template_dir = 'multi_pyramid_wsgi'
    summary = 'Create a multiple WSGI proyect' \
        ' with a Pyramid wsgi application.'


class SimpleGObjTemplate(GinsFSMTemplate):
    _template_dir = 'simple_gobj'
    summary = 'Create a simple GObj proyect.'


class SimpleWsgiTemplate(GinsFSMTemplate):
    _template_dir = 'simple_wsgi'
    summary = 'Create a simple WSGI proyect.'


class SimplePyramidTemplate(GinsFSMTemplate):
    _template_dir = 'simple_pyramid'
    summary = 'Create a simple Pyramid proyect, with sockjs server activated.'


class SimpleWebsocketServerTemplate(GinsFSMTemplate):
    _template_dir = 'simple_websocket_server'
    summary = 'Create a simple Websocket Server proyect'




# -*- encoding: utf-8 -*-
"""
GObj :class:`GTimer`
====================

GObj for generate timeout events.

.. autoclass:: GTimer
    :members:
"""

from ginsfsm.gobj import GObj


def ac_set_timer(self, event):
    self.overwrite_few_parameters(['seconds', 'autostart'], **event.kw)
    if self._timer_id:
        # Clear the current timer
        self.gaplic._clearTimeout(self._timer_id)
        self._timer_id = None

    if self.config.seconds != -1:
        self._timer_id = self.gaplic._setTimeout(
            self.config.seconds,
            self._got_timer,
            autostart=self.config.autostart)


TIMER_FSM = {
    'event_list': (
        'EV_SET_TIMER: top,input',
        'EV_TIMEOUT: top,output',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_SET_TIMER', ac_set_timer, None),
        ),
    }
}

TIMER_GCONFIG = {
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
        " Default is ``None``, i.e., the parent"
    ],
    'timeout_event_name': [
        str, 'EV_TIMEOUT', 0, None, "Timeout event name"
    ],
    'seconds': [
        int, -1, 0, None,
        "timeout in seconds."
        " With ``seconds=-1`` the current timer is cancelled"
    ],
    'autostart': [
        bool, False, 0, None,
        "If ``True`` then timeout event is repeated every ``seconds``"
    ],
}


class GTimer(GObj):
    """  Supply timer events.

    Broadcast the ``timeout_event_name`` event
    when the ``seconds`` time has elapsed.

    If ``autostart`` is ``True``, the timer is cyclic.

    .. ginsfsm::
       :fsm: TIMER_FSM
       :gconfig: TIMER_GCONFIG

    *Input-Events:*

        * :attr:`'EV_SET_TIMER'`: Set the timer.
            Broadcast the ``'EV_TIMEOUT'`` or ``timeout_event_name`` event
            when the ``seconds`` time has elapsed.

            If ``autostart`` is true, the timer is cyclic.

            Associated data to event:

            * ``seconds``: seconds of timer.
            * ``autostart``: cyclic timer.

            With ``seconds=-1`` the current timer is cancelled.

    *Output-Events:*

        * :attr:`'EV_TIMEOUT'`: Timer over.
            Send the ``'EV_TIMEOUT'`` or ``timeout_event_name`` event
            to subscribers.

    """

    def __init__(self):
        GObj.__init__(self, TIMER_FSM, TIMER_GCONFIG)
        self._timer_id = None

    def start_up(self):
        """
        Subcribe the *output-events* to ``subscriber``.

        The output-events are broadcasting using ``post_event()``,
        forcing to get timeout events by queue, not directly.
        """
        if self.config.subscriber is None:
            self.config.subscriber = self.parent
        # force use post_event(), to get timeout events by queue, not directly
        self.subscribe_event(
            None,
            self.config.subscriber,
            __use_post_event__=True
        )

    def _got_timer(self):
        """ callback timer
        """
        self.broadcast_event(self.config.timeout_event_name)



"""
Simple WSGI server
==================

It uses :class:`ginsfsm.protocols.wsgi.server.c_wsgi_server.GWsgiServer`.

You can run this file with ``gserve simple_wsgi_server.ini``

.. autofunction:: main

"""

import logging
logging.basicConfig(level=logging.DEBUG)

from ginsfsm.gaplic import GAplic
from ginsfsm.globals import (
    set_global_app,
    set_global_main_gaplic,
)
from ginsfsm.protocols.wsgi.server.c_wsgi_server import GWsgiServer


#===============================================================
#                   Wsgi app
#===============================================================
def application(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    yield b'Hello World\n'


def paste_app_factory(global_config, **local_conf):
    return application


#===============================================================
#                   Main
#===============================================================
def main(global_config, **local_conf):
    """ Entry point to run with gserve (PasteDeploy)
    """
    if 'application' in local_conf:
        application = local_conf.pop('application')
    else:
        raise Exception('You must supply an wsgi application.')
    ga = GAplic(name='Wsgi-Example', roles='', **local_conf)
    set_global_main_gaplic(ga)
    ga.create_gobj(
        'wsgi-server',
        GWsgiServer,
        ga,
        application=application,
    )
    return ga


if __name__ == "__main__":
    local_conf = {
        'GObj.trace_mach': True,
        'GObj.logger': logging,
        'GSock.trace_dump': True,
        'wsgi-server.host': '0.0.0.0',
        'wsgi-server.port': 8002,
        'application': 'wsgi-application',
    }
    set_global_app(
        'wsgi-application',
        paste_app_factory({}, **local_conf)
    )
    ga = main({}, **local_conf)

    try:
        ga.start()
    except (KeyboardInterrupt, SystemExit):
        print('Program stopped')



# -*- encoding: utf-8 -*-
""" basehandler
"""

import random
import hashlib
from pyramid.view import view_config

from ginsfsm.compat import bytes_
from ginsfsm.protocols.sockjs.server.proto import json_encode
from ginsfsm.protocols.sockjs.server.basehandler import (
    PreflightHandler,
    BaseHandler,
)
from ginsfsm.gobj import GObj
from ginsfsm.gconfig import GConfig
from ginsfsm.protocols.sockjs.server.util import (
    str_to_bytes,
    MAXSIZE,
)

#----------------------------------------------------------------#
#                   Info GClass
#----------------------------------------------------------------#
GINFO_GCONFIG = {
    'sockjs_server': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ""],
}


class GInfo(GObj):
    """  GInfo GObj.
    """
    def __init__(self):
        GObj.__init__(self, {}, GINFO_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        # Info Test views
        """
        self.sockjs_server.pyramid_config.add_view(
            context=GInfo,
            name='',
            view=InfoHandler,
            attr='get',
            path_info=self.resource_path(),
            request_method='GET',
            permission=self.sockjs_server.permission,
        )
        self.sockjs_server.pyramid_config.add_view(
            context=GInfo,
            name='',
            view=InfoHandler,
            attr='options',
            path_info=self.resource_path(),
            request_method='OPTIONS',
            permission=self.sockjs_server.permission,
        )
        """


#----------------------------------------------------------------#
#                   Info views
#----------------------------------------------------------------#
@view_config(
    context=GInfo,
    name='',
    attr='get',
    request_method='GET',
)
@view_config(
    context=GInfo,
    name='',
    attr='options',
    request_method='OPTIONS',
)
class InfoHandler(PreflightHandler):
    """SockJS 0.2+ /info handler"""
    def __init__(self, context, request):
        PreflightHandler.__init__(self, context, request)
        self.access_methods = 'OPTIONS, GET'

    def get(self):
        response = self.response
        response.content_type = 'application/json'
        response.charset = 'UTF-8'
        self.preflight()
        self.disable_cache()

        disabled_transports = \
            self.context.sockjs_server.config.disabled_transports

        info = {
            'websocket': 'websocket' not in disabled_transports,
            'cookie_needed': self.context.sockjs_server.cookie_needed,
            'origins': ['*:*'],
            'entropy': random.randint(0, MAXSIZE),
        }
        response.body = bytes_(json_encode(info))
        return response


#----------------------------------------------------------------#
#                   IFrame GClass
#----------------------------------------------------------------#
GIFRAME_GCONFIG = {
    'sockjs_server': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ""],
}


class GIFrame(GObj):
    """  GIFrame GObj.
    """
    def __init__(self):
        GObj.__init__(self, {}, GIFRAME_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        """
        self.sockjs_server.pyramid_config.add_view(
            context=GIFrame,
#            name='',
            view=IFrameHandler,
            attr='get',
            path_info=self.resource_path(),
            request_method='GET',
            permission=self.sockjs_server.permission,
        )
        """

#----------------------------------------------------------------#
#                   IFrame views
#----------------------------------------------------------------#
IFRAME_TEXT = """<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function(){SockJS.bootstrap_iframe();};
  </script>
  <script src="%s"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>
""".strip()


@view_config(
    context=GIFrame,
    name='',
    attr='get',
    request_method='GET',
)
class IFrameHandler(BaseHandler):
    """SockJS IFrame page handler"""
    def get(self):
        response = self.response
        data = str_to_bytes(
            IFRAME_TEXT % self.context.sockjs_server.config.sockjs_url)
        hsh = hashlib.md5(data).hexdigest()

        #value = self.request.headers.get('If-None-Match')
        value = self.request.environ.get('HTTP_IF_NONE_MATCH')
        if value:
            if value.find(hsh) != -1:
                del response.headers['Content-Type']
                self.set_status(304)
                return response

        response.content_type = 'text/html'
        response.charset = 'UTF-8'
        self.enable_cache()
        self.set_header('Etag', hsh)
        response.body = data
        return response



# -*- encoding: utf-8 -*-

import os
import sys
import ginsfsm

from setuptools import setup, find_packages

if sys.version_info[:2] < (2, 6):
    raise RuntimeError('Requires Python 2.6 or better')

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

setup(
    name='ginsfsm',
    version=ginsfsm.__version__,
    description="GinsFSM, a library to develop systems based in "
                "finite-state machines. It's a communication framework, "
                "including a full asynchronous "
                "http/wsgi/winsocket/sockjs server."
                " Ideal for working with Pyramid using traversal dispatch.",
    long_description=README + "\n\n",  # + CHANGES,
    classifiers=[
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Communications",
        "Topic :: System :: Networking",
        "License :: OSI Approved :: MIT License",
    ],

    keywords='framework communication fsm finite state machine fsm wsgi http '
             'socket poll epoll kqueue select server client workflow pyramid '
             'websocket sockjs',
    author='Ginés Martínez',
    author_email='ginsmar at artgins.com',
    url='http://ginsfsm.org',
    license='MIT License',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'PasteDeploy >= 1.5.0',  # py3 compat
        'Pyramid >= 1.4',  # py3 compat
        #'ujson',
    ],
    extras_require=dict(
        rst=[
            'docutils',
            'svgwrite',
            'pillow',
        ],
    ),
    tests_require=[],
    test_suite="ginsfsm.tests",
    entry_points="""
        [ginsfsm.scaffold]
        multiple_wsgi=ginsfsm.scaffolds:MultipleWsgiTemplate
        multi_pyramid_wsgi=ginsfsm.scaffolds:MultiplePyramidWsgiTemplate
        simple_gobj=ginsfsm.scaffolds:SimpleGObjTemplate
        simple_wsgi=ginsfsm.scaffolds:SimpleWsgiTemplate
        simple_pyramid=ginsfsm.scaffolds:SimplePyramidTemplate
        simple_websocket_server=ginsfsm.scaffolds:SimpleWebsocketServerTemplate

        [console_scripts]
        gcreate = ginsfsm.scripts.gcreate:main
        gserve = ginsfsm.scripts.gserve:main
        gconsole = ginsfsm.scripts.gconsole:main
    """,
)



# -*- encoding: utf-8 -*-
"""
Introduction
============

An :term:`GObj` is an object that has an inside :term:`simple-machine`
that defines its behavior.

The communication between :term:`gobj`'s happens via :term:`event`'s:

:class:`GObj` has the next methods:

* :meth:`GObj.start_up`
* :meth:`GObj.create_gobj`
* :meth:`GObj.destroy_gobj`
* :meth:`GObj.send_event`
* :meth:`GObj.post_event`
* :meth:`GObj.post_event_from_outside`
* :meth:`GObj.broadcast_event`
* :meth:`GObj.subscribe_event`
* :meth:`GObj.delete_subscription`
* :meth:`GObj.set_owned_event_filter`
* :meth:`GObj.overwrite_parameters`
* :meth:`GObj.overwrite_few_parameters`

And this inherited from :class:`ginsfsm.smachine.SMachine`:

* :meth:`ginsfsm.smachine.SMachine.set_new_state`
* :meth:`ginsfsm.smachine.SMachine.get_current_state`
* :meth:`ginsfsm.smachine.SMachine.get_event_list`
* :meth:`ginsfsm.smachine.SMachine.get_output_event_list`


:class:`GObj` has the next public attributes:

* :attr:`GObj.name`
* :attr:`GObj.parent`
* :attr:`GObj.dl_childs`

Events to manage events:

* sending events:

    * sending events by direct delivery: :meth:`GObj.send_event`.
    * sending events by queues: :meth:`GObj.post_event`.
    * sending events from outside world: :meth:`GObj.post_event_from_outside`.
    * sending events to subscribers: :meth:`GObj.broadcast_event`.


* receiving events:

    * directly from another :term:`gobj`'s who knows you.
    * subscribing to events by the :meth:`GObj.subscribe_event` method.
    * you can filtering events being broadcasting with
      :meth:`GObj.set_owned_event_filter` method.


Event
=====

.. autoclass:: Event
    :members:


GObj
====

.. autoclass:: GObj
    :members: start_up, create_gobj, destroy_gobj,
        send_event, post_event, broadcast_event, post_event_from_outside,
        subscribe_event, delete_subscription, set_owned_event_filter,
        overwrite_parameters, overwrite_few_parameters

    .. attribute:: name

        My name.

        Set by :meth:`create_gobj`.

    .. attribute:: parent

        My parent, destination of my events... or not.

        Set by :meth:`create_gobj`.

    .. attribute:: dl_childs

        Set of my gobj childs. Me too like be parent!.


    .. method:: set_new_state

        Set a new state.
        Method to used inside actions, to force the change of state.

        :param new_state: new state to set.

        ``new_state`` must match some of the state names of the
        machine's :term:`state-list` or a :exc:`StateError` exception
        will be raised.

    .. method:: get_current_state

        Return the name of the current state.

        If there is no state it returns ``None``.

    .. method:: get_event_list

       Return the list with the :term:`input-event`'s names.

    .. method:: get_output_event_list

       Return the list with the :term:`output-event`'s names.


Exceptions
==========

.. autoexception:: GObjError

.. autoexception:: EventError

.. autoexception:: StateError

.. autoexception:: MachineError

.. autoexception:: EventNotAcceptedError

"""
import re
import ginsfsm.globals  # made it import available
from ginsfsm.compat import string_types
from ginsfsm.smachine import (
    SMachine,
    EventError,
    EventNotAcceptedError,  # made it import available
    StateError,  # made it import available
    MachineError,  # made it import available
)
from ginsfsm.gconfig import (
    GConfig,
    add_gconfig,
)
from ginsfsm.deferred import Deferred
from ginsfsm.globals import global_get_gobj


class GObjError(Exception):
    """ General GObj exception"""


class Event(object):
    """ Collect event properties. This is the argument received by actions.

    :param destination: destination gobj whom to send the event.
    :param event_name: event name.
    :param source: list with path of gobj sources. Firt item ``source[0]``
        is the original sender gobj. Last item ``source[-1]`` is the
        nearest sender gobj.
    :param kw: keyword arguments with associated data to event.
    """
    # For now, event_factory is private. Automatically using by send_event...
    #Use the :meth:`GObj.event_factory` factory function to create Event
    #instances.
    def __init__(self, destination, event_name, source, **kw):
        self.destination = destination
        self.event_name = event_name
        if not isinstance(source, (list, tuple)):
            source = [source]
        self.source = source
        self.kw = kw.copy()
        self.__dict__.update(**kw)

    def copy(self):
        pass

    def __repr__(self):
        return 'Event object: name %r, destination: %r, kw: %r' % (
            self.event_name,
            self.destination,
            self.kw,
        )

    def __str__(self):
        return 'Event object: name %r, destination: %r, kw: %r' % (
            self.event_name,
            self.destination,
            self.kw,
        )


class _Subscription(object):
    """ Collect subscriber properties
    `event_name`: event name.
    `subscriber_gobj`: subcriber gobj to sending event.
    `kw`: event parameters
    """
    def __init__(self, me, event_name, subscriber_gobj, **kw):
        self.me = me
        self.event_name = event_name
        self.subscriber_gobj = subscriber_gobj
        self.kw = kw
        self.__dict__.update(**kw)

    def __repr__(self):
        resp = "Subscription:\n  me : %r\n  ev : %r\n  sub: %r\n" % (
            self.me,
            self.event_name,
            self.subscriber_gobj)
        return resp


GOBJ_GCONFIG = {
    'trace_traverse': [bool, False, 0, None, 'Trace traverse search'],
    'trace_creation': [
        bool, False, GConfig.FLAG_DIRECT_ATTR, None,
        'Trace creation/destroy of gobj'],
    'trace_subscription': [bool, False, 0, None, 'Trace event subscribing'],
    're_name': [
        str, None, 0, None,
        'Regular expression name to search the gobj in the resource tree.'
        'Used in Pyramid traversal.'],

    # trace_mach is inherited from SMachine.
    'trace_mach': [
        bool, False, GConfig.FLAG_DIRECT_ATTR, None,
        'Trace machine activity'
    ],
    # logger is inherited from SMachine.
    'logger': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],

    'gaplic': [None, None, GConfig.FLAG_DIRECT_ATTR, None, 'My grand-father'],
    'create_gobj': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],
    'destroy_gobj': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],
    '_increase_inside': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],
    '_decrease_inside': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],
    '_tab': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ''],
    '__unique_name__': [
        bool, False, GConfig.FLAG_DIRECT_ATTR, None,
        'If using gaplic, unique name'
    ],
}

_urandom_name = 0


class GObj(SMachine, GConfig):
    """ Well, yes, I'm a very simple brain. Only a machine.
    But write a good FSM, and I never fail you. Derive me, and write my FSM.

    Sample GObj::

        class MyGObj(GObj):
            def __init__(self):
                GObj.__init__(self, FSM, GCONFIG)

            def start_up(self):
                ''' Initialization zone.'''

    :param fsm: FSM :term:`simple-machine`.
    :param gconfig: GCONFIG :term:`gconfig-template`.
    """

    # name and parent attributes must be Location-Aware Resources
    # for Pyramid framework compatibility:
    # changed to point to __name__ and __parent__.
    @property
    def name(self):
        return self.__name__

    @name.setter
    def name(self, name):
        self.__name__ = name

    @property
    def parent(self):
        return self.__parent__

    @parent.setter
    def parent(self, parent):
        self.__parent__ = parent

    def __init__(self, fsm, gconfig=None):
        SMachine.__init__(self, fsm)

        self.name = ''
        """ My name.
        Set by :meth:`create_gobj`
        """
        self.parent = None
        """My parent, destination of my events... or not.
        Set by :meth:`create_gobj`
        """
        self.dl_childs = set()        # my childs... me too like be parent.
        """List of gobj childs.
        """
        self.owned_event_filter = None  # TODO debe ser una lista de filtros
        """Filter to broadcast_event function to check the owner of events.
        """
        self._dl_subscriptions = set()      # uauuu, how many fans!!
        self._some_subscriptions = False
        self._destroyed = False  # mark as destroyed when destroy_gobj()
        self._re_compiled_name = ''  # re compiled name when using re_name
        self.re_matched_name = ''  # matched name when using re_name
        self.trace_creation = False
        gconfig = add_gconfig(gconfig, GOBJ_GCONFIG)
        GConfig.__init__(self, gconfig, self.logger)

    def __del__(self):
        if self.trace_creation:
            self.logger and self.logger.info("Destroyed! <-- %r" % (self))

    def __str__(self):
        name = "%s" % (self.name)
        parent = self.parent
        while(parent):
            parent_name = "%s" % (parent.name)
            name = parent_name + '.' + name
            parent = parent.parent
        return "%s" % (name)

    def __repr__(self):
        name = "%s:%s" % (self.__class__.__name__, self.name)
        parent = self.parent
        while(parent):
            parent_name = "%s:%s" % (parent.__class__.__name__, parent.name)
            name = parent_name + '.' + name
            parent = parent.parent
        return "'%s: %s'" % (name, 'Destroyed' if self._destroyed else "Lived")

    def create_gobj(self, name, gclass, parent, **kw):
        """ Factory function to create gobj's instances.

        :param name: Name of the gobj.

        :param gclass: `gclass` is the :class:`GObj` type used to create
            the new gobj. It's must be a derived class of :class:`GObj`
            otherwise a :exc:`GObjError` exception will be raised.

        :param parent: parent of the new :term:`gobj`. ``None`` if it has no
            parent. If it's not ``None``, then must be a derived class
            of :class:`GObj` otherwise a :exc:`GObjError`
            exception will be raised.

        :param kw: Attributes that are added to the new :term:`gobj`.
            All the keyword arguments used in the creation function
            **are added as attributes** to ``config`` object.

            You must consult the attributes supported
            by each `gclass` type.
            The attributes must be defined in the gclass GCONFIG,
            otherwise they are ignored.

            Special kw:

                ``'__random_name__'``: Generate a random gobj name

            TODO: doc re_name

        :rtype: new gobj instance.

        The factory funcion does:

        * Add the :term:`gobj` to their parent child list
          :attr:`GObj.dl_childs`,
        * If it's a :term:`named-gobj` call the
          :meth:`GObj.register_unique_gobj`.
        * Call :meth:`GObj.start_up`.
        * Add to the :term:`gobj` several attributes:

            * **name**: name of the created :term:`gobj`.
            * **parent**: the parent :term:`gobj` of the created :term:`gobj`.
        """

        if gclass is None:
            raise GObjError(
                '''ERROR create_gobj(): No GObj class supplied.''')
        if not issubclass(gclass, GObj):
            raise GObjError(
                '''ERROR create_gobj(): class '%r' is NOT a GObj subclass''' %
                (gclass))

        if parent is not None:
            if not isinstance(parent, GObj):
                raise GObjError(
                    '''ERROR create_gobj(): parent '%r' '''
                    '''is NOT a GObj subclass''' % (gclass))

        self.__random_name__ = kw.pop('__random_name__', False)
        if self.__random_name__:
            name = self.get_random_name(name)

        gobj = gclass()

        if name:
            gobj.name = name

        # Who wins? arguments or file ini_settings?
        # ini global (in gaplic) wins.
        gobj.write_parameters(**kw)
        if self.gaplic and self.gaplic.ini_settings is not None:
            # ini global win.
            gobj.overwrite_parameters(0, **self.gaplic.ini_settings)

        if parent is not None:
            parent._add_child(gobj)

        if gobj.config.re_name:
            gobj._re_compiled_name = re.compile(gobj.config.re_name)

        if gobj.trace_creation:
            self.logger and self.logger.info("Creating --> '%s:%s'" % (
                gclass.__name__, name))

        if gobj.__unique_name__:
            registered = self.gaplic._register_unique_gobj(gobj)
            if not registered:
                raise GObjError(
                    "ERROR create_gobj():"
                    " cannot _register_unique_gobj '%s' " % (gobj.name))

        gobj.start_up()

        if gobj.trace_creation and self.logger:
            if gobj.__unique_name__:
                unique = '!'
            else:
                unique = ' '
            self.logger.info("Created <--%s '%r'" % (unique, gobj))

        return gobj

    def get_random_name(self, prefix):
        global _urandom_name
        _urandom_name += 1
        if prefix:
            return '%s_%d' % (prefix, _urandom_name)
        else:
            return '%d' % (_urandom_name)

    @staticmethod
    def destroy_gobj(gobj):
        """ Destroy a gobj
        """
        if gobj.trace_creation and gobj.logger:
            gobj.logger.info("Destroying --> %r" % (gobj))

        if gobj._destroyed:
            if gobj.logger:
                gobj.logger.error(
                    "ERROR reentering in destroy_gobj: %r" % gobj)
            return

        gobj._destroyed = True
        if gobj.parent is not None:
            gobj.parent._remove_child(gobj)

        while len(gobj.dl_childs):
            try:
                for child in gobj.dl_childs:
                    if not child._destroyed:
                        GObj.destroy_gobj(child)
            except RuntimeError:
                pass  # "Set changed size during iteration" is OK

        gobj.delete_all_subscriptions(force=True)
        gobj.go_out()
        del gobj

    def start_up(self):
        """ Initialization zone.

        Well, the __init__ method is used to build the FSM so I need another
        function to initialize the new gobj.
        Please, **override me**, and write here all the code you need to
        start up the machine: create your owns childs, etc.
        This function is called by :meth:`create_gobj`
        after creating the gobj instance.
        """

    def go_out(self):
        """ Finish zone.

        Please, **override me** to do extra work
        when the gobj is being destroyed.

        In this point, all childs and subscriptions are already deleted.
        """

    def _resolv_destination(self, destination):
        """ Resolve the destination gobj.
            If destination it's a string:
                * check is gaplic exists.
                * try to resolv the destination by gaplic.

                If the string destination cannot be resolved
                then return the string. It must be resolved by gaplic router.
        """
        if not (isinstance(destination, (string_types, GObj))):
            raise GObjError(
                'Destination gobj ("%s") must be a string or a GObj instance'
                % (destination)
            )

        if isinstance(destination, string_types):
            if not self.gaplic:
                raise GObjError(
                    'When destination gobj ("%s") is a string, '
                    'you need a GAplic!' % (destination)
                )

            named_gobj = self.gaplic.find_unique_gobj(destination)
            if not named_gobj:
                raise GObjError('GObj %r UNKNOWN in %r' % (
                    destination, self.gaplic))
            destination = named_gobj

        return destination

    def _event_factory(self, destination, event, **kw):
        """ Factory to create Event instances.

        :param destination: destination gobj whom send the event.
        :param event: an :term:`event`.
        :param kw: keyword arguments with associated data to event.
        :rtype: Return Event instance.

        ``event`` must be a `string` or :class:`Event` types, otherwise a
        :exc:`EventError` will be raised.

        If ``event`` is an :class:`Event` instance, a new :class:`Event`
        duplicated instance is returned, but it will be updated with
        the new ``destination`` and ``kw`` keyword arguments.

        .. note::
            All the keyword arguments used in the factory function
            **are added as attributes** to the created :term:`event` instance.
            You must consult the attributes supported by each machine's event.
        """
        if not (isinstance(event, string_types) or
                isinstance(event, Event)):
            raise EventError(
                'Event "%r" must be a string or Event instance' %
                (event,))

        # if destination is not None:
        if not (isinstance(destination, string_types) or
                isinstance(destination, GObj) or
                isinstance(destination, Deferred)
                ):
            raise GObjError(
                'Destination "%r" must be a string or GObj/Deferred instance' %
                (destination,))

        if isinstance(event, Event):
            # duplicate the event
            if event.source[-1] != self:
                event.source.append(self)
            if len(kw):
                event.kw.update(**kw)
            event = Event(
                event.destination,
                event.event_name,
                event.source,
                **event.kw
            )
            if destination is not None:
                event.destination = destination
        else:
            event = Event(destination, event, self, **kw)

        return event

    def send_event(self, destination, event, **kw):
        """
        Send ``event`` to ``destination``, with associated event data ``kw``.

        :param destination: Must be a string gobj name or
            a :term:`gobj` instance,
            otherwise a :exc:`GObjError` will be raised.
            If it's a string and the gobj name cannot be resolved
            inside the current :term:`gaplic`, then the event is
            :func:`post_event` in order to be resolved by the gaplic router.

        :param event: Must be a string event name or a :class:`Event` instance,
            otherwise an :exc:`EventError` will be raised.

        :param kw: All the keyword arguments **are added as attributes** to
            the sent :class:`Event` instance.
            You must consult the attributes supported by each machine's event.

        :rtype: return the returned value from the executed action.
            Return ``None`` if the event has been :func:`post_event`.

        If the :term:`event-name` exists in the machine, but it's not accepted
        by the current state, then no exception is raised but the
        function **returns** :exc:`EventNotAcceptedError`.

        .. note:: The :meth:`inject_event` method doesn't
            **raise** :exc:`EventNotAcceptedError` because a
            :term:`machine` should run under any circumstances.
            In any way an action can raise exceptions.

        """
        destination = self._resolv_destination(destination)
        event = self._event_factory(destination, event, **kw)

        if isinstance(destination, string_types):
            # string gobjs must be resolved by gaplic
            return self.post_event(destination, event)

        if destination._destroyed:
            self.logger and self.logger.error(
                "ERROR internal: sending event %r to a destroyed gobj %r" % (
                    event.event_name, destination)
            )
            return -1

        ret = destination.inject_event(event)
        return ret

    @staticmethod
    def post_event_from_outside(gaplic_name, gobj_name, event, **kw):
        """ Same funcionality as :func:`post_event`,
        but the event is sent from non-gobj world.
        """
        gaplic = ginsfsm.globals.global_get_gaplic(gaplic_name)
        if not gaplic:
            raise Exception("CHECK NAMES!! gaplic %s  doesn't exist" % (
                gaplic_name,))
        gobj = global_get_gobj(gobj_name, gaplic_name)
        if not gobj:
            raise Exception("CHECK NAMES!! gobj %s  doesn't exist" % (
                gobj_name,))
        gaplic.post_event(gobj, event, **kw)

    def post_event(self, destination, event, **kw):
        """ Same funcionality as :func:`send_event`,
        but the event is processed by gaplic.
        If the destination is inside of the current gaplic the event will be
        sent in the next loop cycle.
        """
        event = self._event_factory(destination, event, **kw)
        if not self.gaplic:
            raise GObjError(
                'To use post_event, you need a GAplic!'
            )
        self.gaplic.enqueue_event(event)

    def broadcast_event(self, event, **kw):
        """ Broadcast the ``event`` to all subscribers.

        :param event: :term:`event` to send.
        :param kw: keyword argument with data associated to event.

            .. note::
                All the keyword arguments **are added as attributes** to
                the sent :term:`event`.

        Use this function when you don't know who are your event's clients,
        when you don't know the :term:`gobj` destination of
        your :term:`output-event`'s

        If there is no subscriptors, the event is not sent.

        When an event has several subscriptors, there is a mechanism called
        :term:`event-filter` that allows to a subcriptor to own the event
        and no further spread by more subscribers.

        The filter function set by :meth:`set_owned_event_filter` method,
        is call with the returned value of an :term:`action` as argument:
        If the filter function return ``True``, the event is owned, and the
        :func:`ginsfsm.gobj.GObj.broadcast_event` function doesn't continue
        sending the event to other subscribers.

        .. note:: If :func:`ginsfsm.gobj.GObj.broadcast_event` function
           uses :func:`ginsfsm.gobj.GObj.post_event`,
           the :term:`event-filter` cannot be applied.
        """
        if self.config.trace_subscription and self.logger:
            self.logger.info('BROADCAST event: %r, gobj: %r' % (event, self))

        if not self._some_subscriptions:
            return
        subscriptions = self._dl_subscriptions.copy()
        sended_gobj = set()  # don't repeat events
        for sub in subscriptions:
            if sub.subscriber_gobj in sended_gobj:
                continue

            oevent = self._event_factory(sub.subscriber_gobj, event, **kw)
            if None in sub.event_name or \
                    oevent.event_name in sub.event_name:

                if hasattr(sub, '__rename_event_name__'):
                    oevent.kw.update(
                        {
                            'original_event_name': oevent.event_name,
                        }
                    )
                    oevent.event_name = sub.__rename_event_name__

                if hasattr(sub, '__subscription_reference__'):
                    oevent.kw.update(
                        {
                            '__subscription_reference__':
                            sub.__subscription_reference__,
                        }
                    )
                    oevent.event_name = sub.__rename_event_name__

                ret = False
                if isinstance(sub.subscriber_gobj, Deferred):
                    # outside world
                    if hasattr(sub, '__deferred_witout_oevent__'):
                        sub.subscriber_gobj()
                    else:
                        sub.subscriber_gobj(event=oevent)
                else:
                    # gobj-ecosistema
                    if hasattr(sub, '__use_post_event__'):
                        ret = self.post_event(sub.subscriber_gobj, oevent)
                    else:
                        ret = self.send_event(sub.subscriber_gobj, oevent)
                        if self.owned_event_filter:
                            ret = self.owned_event_filter(ret)
                            if ret is True:
                                return  # propietary event
                sended_gobj.add(sub.subscriber_gobj)

    def subscribe_event(self, event_name, subscriber_gobj, **kw):
        """ Subscribe to an event.

        :param event_name: string event name or tuple/list of string
            event names.  If ``event_name`` is ``None`` then it subscribes
            to all events. If it's not ``None`` then it must be a valid event
            name from the :term:`output-event` list,
            otherwise a :exc:`EventError` will be raised.

        :param subscriber_gobj: subscriber obj that wants receive the event.

            ``subscriber_gobj`` must be:
                * `None`: the subscriber is the parent.
                * `string`: the subscriber is a :term:`unique-named-gobj`.
                * type :class:`GObj` instance.
                * Deferred callback.

            otherwise otherwise a :exc:`GObjError` will be raised.

        :param kw: keyword arguments.
        :return: the subscription object.

        Possible values for **kw** arguments:
            * `__use_post_event__`: ``bool``

              You must set it to `True` in order to broadcast the events
              using `post-event` instead of `send-event`.

            * `__rename_event_name__`: `'new event name'`

              You can rename the output original event name.
              The :attr:`original_event_name` attribute is added to
              the sent event with the value of the original event name.


            * `__hard_subscription__`:  ``bool``

              True for permanent subscription.

              This subscription cannot be remove,
              neither with delete_all_subscriptions().
              (Well, with force you can)

            * `__subscription_reference__`:  ``str``

              If exists, it will be added as kw in the event broadcast.
              Can be used by the subscriptor for general purposes.

            * `__deferred_witout_oevent__`:  ``Bool``

              If True, the callback is called without event keyword parameter.
        """
        #if subscriber_gobj is not None:
        if not (isinstance(subscriber_gobj, string_types) or
                isinstance(subscriber_gobj, Deferred) or
                isinstance(subscriber_gobj, GObj)
                ):
            raise GObjError('Bad type of subscriber_gobj %r' % (
                subscriber_gobj,))

        output_events = self.get_output_event_list()

        if not isinstance(event_name, (list, tuple)):
            event_name = (event_name,)

        for name in event_name:
            if name is None:
                continue
            if not isinstance(name, string_types):
                raise EventError(
                    'subscribe_event(): event %r is not string in %r'
                    % (name, self))

            if name not in output_events:
                raise EventError(
                    'subscribe_event(): output-event %r not defined in'
                    ' %r' % (event_name, self))

        existing_subs = self._find_subscription(event_name, subscriber_gobj)
        if existing_subs:
            # avoid duplication subscriptions
            if self.logger:
                self.logger.warning(
                    "WARNING duplicate subscription:\n"
                    "  me : %r\n"
                    "  ev : %r\n"
                    "  sub: %r\n" %
                    (self, event_name, subscriber_gobj))
            self.delete_subscription(event_name, subscriber_gobj)
        subscription = _Subscription(self, event_name, subscriber_gobj, **kw)
        if self.config.trace_subscription and self.logger:
            self.logger.info('NEW %r' % subscription)
        self._dl_subscriptions.add(subscription)
        self._some_subscriptions = True
        return subscription

    def _find_subscription(self, event_name, subscriber_gobj):
        """ Find a subscription by event_name and subscriber gobj.
        Internal use to avoid duplicates subscriptions.
        """
        if not isinstance(event_name, (list, tuple)):
            event_name = (event_name,)
        for sub in self._dl_subscriptions:
            if sorted(list(sub.event_name)) == sorted(list(event_name)):
                if sub.subscriber_gobj == subscriber_gobj:
                    return sub
        return None

    def delete_subscription_by_object(self, subscription):
        if subscription in self._dl_subscriptions:
            if self.config.trace_subscription and self.logger:
                self.logger.info('DEL %r' % subscription)
            self._dl_subscriptions.remove(subscription)
            if len(self._dl_subscriptions) == 0:
                self._some_subscriptions = False
            return True
        return False

    def delete_subscription(self, event_name, subscriber_gobj):
        """ Remove `subscription`.

        :param event_name: string event name or tuple/list of string
            event names.
        :param subscriber_gobj: subscriber gobj.
        """
        sub = self._find_subscription(event_name, subscriber_gobj)
        if self.config.trace_subscription and self.logger:
            self.logger.info('DEL %r' % sub)
        if sub:
            if sub.kw.get('__hard_subscription__', None):
                if self.logger:
                    self.logger.error(
                        "WARNING cannot delete a hard subscription(): '%s'" % (
                            event_name))
                return False
            else:
                self._dl_subscriptions.remove(sub)
                if len(self._dl_subscriptions) == 0:
                    self._some_subscriptions = False
                return True

        if self.logger:
            self.logger.error(
                "ERROR delete_subscription(): '%s' NOT FOUND " % (event_name))

        return False

    def delete_all_subscriptions(self, force=False):
        """ Remove all subscriptions.
        """
        subscriptions = self._dl_subscriptions.copy()
        for sub in subscriptions:
            if not force:
                if sub.kw.get('__hard_subscription__'):
                    continue

            if self.config.trace_subscription and self.logger:
                self.logger.info('DEL %r' % sub)
            self._dl_subscriptions.remove(sub)

        if len(self._dl_subscriptions) == 0:
            self._some_subscriptions = False

    def set_owned_event_filter(self, filter):
        """ Set a filter function to be used
        by :meth:`broadcast_event` function to check the owner of events.
        """
        self.owned_event_filter = filter

    def _add_child(self, gobj):
        """ Add a child ``gobj``.

        :param gobj: :term:`gobj` child to add.

        Raise :exc:`GObjError` is ``gobj`` already has a parent.
        This function is called by :meth:`create_gobj`
        after creating the gobj instance.
        """
        if gobj.parent:
            raise GObjError('GObj "%r" already has parent' % (gobj))
        self.dl_childs.add(gobj)
        gobj.parent = self

    def _remove_child(self, gobj):
        """ Remove the child ``gobj``.

        :param gobj: :term:`gobj` child to remove.

        This function is called by :meth:`destroy_gobj`.
        """
        if gobj in self.dl_childs:
            self.dl_childs.remove(gobj)
            gobj.parent = None

    def __getitem__(self, name):
        """ Enable gobjs tree to work with Pyramid traversal URL dispatch.
        """
        trace_traverse = self.config.trace_traverse
        if trace_traverse and self.logger:
            self.logger.debug("==> TRAVERSING('%s') -> %s" % (
                name, self.resource_path()))

        if name is None:
            raise KeyError('Parameter name cannot be None')
        # Firstly search no re_name
        for gobj in self.dl_childs:
            if gobj.config.re_name:
                continue
            else:
                if gobj.name == name:
                    if trace_traverse and self.logger:
                        self.logger.debug("    TRAVERSING FOUND!")
                    return gobj

        # Secondly search re_name
        for gobj in self.dl_childs:
            if gobj.config.re_name:
                if gobj._re_compiled_name.match(name) is not None:
                    # this gobj has been got as re_matched_name.
                    gobj.re_matched_name = name
                    if trace_traverse and self.logger:
                        self.logger.debug("    TRAVERSING %s FOUND RE!" % name)
                    return gobj
            else:
                continue
        if trace_traverse and self.logger:
            self.logger.debug("<== TRAVERSING('%s') -> %s END!" % (
                name, self.resource_path()))
        raise KeyError('No such child named %s' % name)

    def __iter__(self):
        """ Iterates over child elements.
        """
        return self.dl_childs.__iter__()

    # helper for pyramid
    def resource_path(self, separator='/'):
        """ Same as Pyramid traversal.resource_path, but without escaping.
        """
        path = [loc.__name__ or '' for loc in lineage(self)]
        path.reverse()
        return path and separator.join([x for x in path]) or separator

    def overwrite_parameters(self, level, **settings):
        """ The parameters in settings must be defined in the gobj.

        :param level: level trace of childs.
            indicates the depth of the childs as far to change.

            * `0` only this gobj.
            * `-1` all tree of childs.

        :param settings: parameters and their values.

        The settings are filtered by the named-gobj
        or gclass name of this gobj.
        The parameter name in settings, must be a dot-named,
        with the first item being the named-gobj o gclass name.
        """
        parameters = self.filter_parameters(**settings)
        self.write_parameters(**parameters)
        if level > 0:
            for child in self.dl_childs:
                child.overwrite_parameters(level - 1, **settings)

    def overwrite_few_parameters(self, parameter_list, **settings):
        """ The parameters in settings must be defined in the gobj.

        :param parameter_list: write only the parameters in ``parameter_list``.

        :param settings: parameters and their values.

        The settings are filtered by the named-gobj
        or gclass name of this gobj.
        The parameter name in settings, must be a dot-named,
        with the first item being the named-gobj o gclass name.
        """
        parameters = self.filter_parameters(**settings)
        self.write_few_parameters(parameter_list, **parameters)


def inside(resource1, resource2):
    """Is ``resource1`` 'inside' ``resource2``?  Return ``True`` if so, else
    ``False``.

    ``resource1`` is 'inside' ``resource2`` if ``resource2`` is a
    :term:`lineage` ancestor of ``resource1``.  It is a lineage ancestor
    if its parent (or one of its parent's parents, etc.) is an
    ancestor.
    """
    while resource1 is not None:
        if resource1 is resource2:
            return True
        resource1 = resource1.__parent__

    return False


def lineage(resource):
    """
    Return a generator representing the :term:`lineage` of the
    :term:`resource` object implied by the ``resource`` argument.  The
    generator first returns ``resource`` unconditionally.  Then, if
    ``resource`` supplies a ``__parent__`` attribute, return the resource
    represented by ``resource.__parent__``.  If *that* resource has a
    ``__parent__`` attribute, return that resource's parent, and so on,
    until the resource being inspected either has no ``__parent__``
    attribute or which has a ``__parent__`` attribute of ``None``.
    For example, if the resource tree is::

      thing1 = Thing()
      thing2 = Thing()
      thing2.__parent__ = thing1

    Calling ``lineage(thing2)`` will return a generator.  When we turn
    it into a list, we will get::

      list(lineage(thing2))
      [ <Thing object at thing2>, <Thing object at thing1> ]
    """
    while resource is not None:
        yield resource
        # The common case is that the AttributeError exception below
        # is exceptional as long as the developer is a "good citizen"
        # who has a root object with a __parent__ of None.  Using an
        # exception here instead of a getattr with a default is an
        # important micro-optimization, because this function is
        # called in any non-trivial application over and over again to
        # generate URLs and paths.
        try:
            resource = resource.__parent__
        except AttributeError:
            resource = None



# -*- encoding: utf-8 -*-
""" Draw a SMachine using svg format.
"""
import svgwrite
from .layout import Layer
import svgwrite.mixins


class FsmDraw(object):
    """ Draw a SMachine using svg format.
    """
    def __init__(self, smachine, filepath, font_name, font_size):
        self.smachine = smachine
        self.filepath = filepath
        self.font_name = font_name
        self.font_size = font_size
        self.drawer = None

    def draw(self):
        layout = Layer(self.smachine, self.font_name, self.font_size)
        self.render(layout)

    def pagesize(self):
        return self.size

    def save(self, size=None):
        """ save to file
        """
        #----------------------------#
        #   Save and convert to png
        #----------------------------#
        self.drawer.save()
        #data = open(self.filepath).read()
        #svg2png(data, self.filepath + '.png')

    def render(self, layout):
        """ Render the layout to svg
        """
        #--------------------------#
        #   Init drawing
        #--------------------------#
        self.size = layout.size
        self.drawer = drawer = svgwrite.Drawing(
            filename=self.filepath,
            profile='tiny',
            size=self.size,
            debug=False,
        )
        drawer.set_desc(title="", desc="Graph of a FSM")

        #
        #   Add filter_blur to use with all state's node boxes.
        #
        drawer['xmlns:inkspace'] = 'http://www.inkscape.org/namespaces/inkscape'
        # inkspace's Gaussian filter
        filtr = drawer.defs.add(
            drawer.filter(
                id='filter_blur',
                start=(-0.07875, -0.252),  # original -0.07875, -0.252
                size=(1.1575, 1.504),  # original 1.1575, 1.504
            )
        )
        filtr['inkspace:collect'] = 'always'
        gaussian = filtr.feGaussianBlur(
            id='feGaussianBlur3780',
            stdDeviation=4.2)
        gaussian['inkspace:collect'] = 'always'

        #--------------------------#
        #   Add layout
        #--------------------------#
        self.drawer.add(layout)



# -*- encoding: utf-8 -*-
"""
banco
=====

It uses :class:`ginsfsm.protocols.wsgi.server.c_wsgi_server.GWsgiServer`.

.. autofunction:: main

"""

from ginsfsm.gaplic import GAplic
from ginsfsm.globals import (
    set_global_app,
    set_global_main_gaplic,
)
from ginsfsm.protocols.wsgi.server.c_wsgi_server import GWsgiServer
from banco.pyramid_app import pyramid_wsgi_app
from banco.ac_cuentas import Cuentas


#===============================================================
#       Paste app's factory.
#       To run with gserve banco.ini
#===============================================================
def pyramid_paste_factory(global_config, **local_conf):
    """ Returns the WSGI application.
    """
    return pyramid_wsgi_app(global_config, **local_conf)


#===============================================================
#                   Main
#===============================================================
def main(global_config, **local_conf):
    """ Entry point to run with gserve (PasteDeploy)
    """
    if 'application' in local_conf:
        application = local_conf.pop('application')
    else:
        raise Exception('You must supply a wsgi application in local_conf.')

    if 'gaplic_name' in local_conf:
        gaplic_name = local_conf.pop('gaplic_name')
    else:
        raise Exception('You must supply a gaplic name in local_conf.')

    ga = GAplic(
        name=gaplic_name,
        roles=('banco',),
        **local_conf)
    set_global_main_gaplic(ga)
    ga.create_gobj(
        'wsgi_server',
        GWsgiServer,
        ga,
        application=application,
    )
    ga.create_gobj(
        'cuentas',
        Cuentas,
        ga,
        __unique_name__=True,
    )
    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
    print("<<<     To test the application, go to http://localhost:8001   >>>")
    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
    return ga


if __name__ == "__main__":
    """ You can run directly this file as main, without gserve.
    """
    import logging
    logging.basicConfig(level=logging.DEBUG)

    # simulate running from ini file
    local_conf = {
        'gaplic_name': 'main_gaplic',
        'application': 'wsgi-application',
        'GWsgiServer.host': '0.0.0.0',
        'GWsgiServer.port': 8001,
        'GSock.trace_dump': False,
        'GObj.trace_mach': False,
        'GObj.trace_creation': False,
        'GObj.trace_traverse': False,
        'GObj.trace_subscription': False,

        'GObj.logger': logging,
        'PyramidRoot.pyramid_router_url': '__pyramid_router__',
        'GRouter.trace_router': True,

        'GRouter.static_routes':
        'RobberA, financiera+bolsa+publicidad, http://localhost:8002;',
    }

    ga = main({}, **local_conf)

    pyramid_settings = {
        'gaplic': 'main_gaplic',
        'pyramid.reload_templates': True,
        'pyramid.debug_authorization': False,
        'pyramid.debug_notfound': False,
        'pyramid.debug_routematch': False,
        'pyramid.debug_templates': True,
        'pyramid.default_locale_name': 'en',
    }

    set_global_app(
        'wsgi-application',
        pyramid_paste_factory({}, **pyramid_settings)
    )

    try:
        ga.start()
    except (KeyboardInterrupt, SystemExit):
        print('Program stopped')



# -*- encoding: utf-8 -*-

from .utils import get_textsize


class Metrics(object):
    def __init__(self, font_name, font_size):
        # Calculate state box metrics
        # box action zone
        self.font_name = font_name
        self.font_size = font_size
        self.calculate_metrics()

    def calculate_metrics(self):
        # use the A letter as base
        size, ascent, descent = get_textsize(
            "AA", self.font_name, self.font_size)

        # state title
        self.ascent = ascent
        self.descent = descent
        self.y_margin = size.height / 2
        self.x_margin = size.width * 1
        self.states_y_margin = 40

        self.state_change_line_width = 2
        self.event_line_width = 4
        self.tt_axis = 120
        self.t_axis = 50
        self.bb_axis = 50

        self.top_margin = 40
        self.bottom_margin = 30
        self.left_margin = 30
        self.right_margin = 60

        self.ev_line_sep = 25
        self.st_line_sep = 10



"""
GObj :class:`OnGConnex`
=======================

Utility for check :class:`ginsfsm.c_connex.GConnex`.

Send a query to two url's alternatively.

How? send the query each 10 seconds, but set a timeout_inactivity of 5.
In each disconnection the next connection will switch between the
two destinations.

.. autoclass:: OnGConnex
    :members:

"""

import logging
logging.basicConfig(level=logging.DEBUG)

from ginsfsm.gobj import GObj
from ginsfsm.gaplic import GAplic
from ginsfsm.c_timer import GTimer
from ginsfsm.c_connex import GConnex
from ginsfsm.utils import hexdump

QUERY = b"GET / HTTP/1.1\r\n" + \
    b"Host: www\r\n" + \
    b"\r\n"


def ac_rx_data(self, event):
    data = event.kw['data']
    if self.config.verbose:
        print('Receiving %d bytes' % len(data))
        print(hexdump('<=', data))


def ac_timeout(self, event):
    self.send_event(self.connex, 'EV_SEND_DATA', data=QUERY)
    self.set_timeout(10)


ONGCONNEX_FSM = {
    'event_list': (
        'EV_SET_TIMER: bottom output',
        'EV_TIMEOUT:bottom input',
        'EV_RX_DATA:bottom input',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_TIMEOUT',      ac_timeout,     'ST_IDLE'),
            ('EV_RX_DATA',      ac_rx_data,     'ST_IDLE'),
        ),
    }
}


ONGCONNEX_GCONFIG = {  # type, default_value, flag, validate_function, desc
    'verbose': [int, 0, 0, None, "Increase output verbosity. Values [0,1,2]"],
    'url1': [str, 'www.google.com', 0, None, "Url one to connect."],
    'url2': [str, 'www.twitter.com', 0, None, "Url two to connect."],
}


class OnGConnex(GObj):
    """  OnGConnex GObj.

    .. ginsfsm::
       :fsm: ONGCONNEX_FSM
       :gconfig: ONGCONNEX_GCONFIG

    *Input-Events:*

        * :attr:`'EV_TIMEOUT'`: Timer over.
            Start the machine.

        * :attr:`'EV_RX_DATA'`: Receiving data.
            Receiving data from urls.

    *Output-Events:*

        * :attr:`'EV_START_TIMER'`: Start timer.

    """
    def __init__(self):
        GObj.__init__(self, ONGCONNEX_FSM, ONGCONNEX_GCONFIG)

    def start_up(self):
        """ Initialization zone."""
        self.timer = self.create_gobj(
            None,
            GTimer,
            self,
        )

        self.connex = self.create_gobj(
            'connex',
            GConnex,
            self,
            destinations=[
                (self.config.url1, 80),
                (self.config.url2, 80),
            ],
            timeout_inactivity=5,
            # only want receive EV_RX_DATA event
            connected_event_name=None,
            disconnected_event_name=None,
            transmit_ready_event_name=None,
        )

        self.set_timeout(5)

    def set_timeout(self, seconds):
        self.send_event(self.timer, 'EV_SET_TIMER', seconds=seconds)

    def clear_timeout(self):
        self.send_event(self.timer, 'EV_SET_TIMER', seconds=-1)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "url1",
        nargs='?', default='www.google.es',
        help="First url to connect."
    )
    parser.add_argument(
        "url2",
        nargs='?', default='www.twitter.es',
        help="Second url to connect."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase output verbosity",
        type=int,
        choices=[0, 1, 2],
        default=1,
    )
    args = parser.parse_args()

    local_conf = {
        'GObj.trace_mach': True if args.verbose else False,
        'GObj.logger': logging,
    }

    ga = GAplic(name='', roles='', **local_conf)
    ga.create_gobj(
        'ongconnex',
        OnGConnex,
        ga,
        verbose=args.verbose,
        url1=args.url1,
        url2=args.url2,
    )
    try:
        ga.start()
    except (KeyboardInterrupt, SystemExit):
        print('Program stopped')



# -*- encoding: utf-8 -*-
"""
applic
======

TestApps

"""
from pyramid.security import (
    Allow,
    Everyone,
    ALL_PERMISSIONS,
)

from ginsfsm.gobj import GObj
from ginsfsm.c_timer import GTimer
from ginsfsm.protocols.sockjs.server.c_sockjs_server import GSockjsServer

from .test import (
    EchoConnection,
    CloseConnection,
    CookieEcho,
)

#----------------------------------------------------------------------#
#                           TestApps
#----------------------------------------------------------------------#
TESTAPPS_FSM = {}

TESTAPPS_GCONFIG = {
}


class TestSocksjsApps(GObj):
    """  Root resource for TestApps.

    .. ginsfsm::
       :fsm: TESTAPPS_FSM
       :gconfig: TESTAPPS_GCONFIG

    *Input-Events:*
        * :attr:`'EV_TIMEOUT'`: Timer over.

    """
    __acl__ = [
        (Allow, Everyone, ALL_PERMISSIONS)
    ]

    def __init__(self):
        GObj.__init__(self, TESTAPPS_FSM, TESTAPPS_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """

        self.timer = self.create_gobj(
            None,       # unnamed gobj
            GTimer,     # gclass
            self        # parent
        )
        #self.send_event(self.timer, 'EV_SET_TIMER', seconds=self.seconds)

        # create test sockjs apps
        self.echo = self.create_gobj(
            'echo',         # named gobj
            GSockjsServer,  # gclass
            self,           # parent
            sockjs_app_class=EchoConnection,
            response_limit=4096,
        )
        self.disabled_websocket_echo = self.create_gobj(
            'disabled_websocket_echo',
            GSockjsServer,
            self,
            sockjs_app_class=EchoConnection,
            disabled_transports=['websocket'],
        )
        self.disabled_websocket_echo = self.create_gobj(
            'close',
            GSockjsServer,
            self,
            sockjs_app_class=CloseConnection,
        )
        #self.disabled_websocket_echo = self.create_gobj(
        #    'ticker',
        #    GSockjsServer,
        #    self,
        #    sockjs_app_class=TickerConnection,
        #)
        #self.disabled_websocket_echo = self.create_gobj(
        #    'amplify',
        #    GSockjsServer,
        #    self,
        #    sockjs_app_class=AmplifyConnection,
        #)
        self.disabled_websocket_echo = self.create_gobj(
            'cookie_needed_echo',
            GSockjsServer,
            self,
            sockjs_app_class=CookieEcho,
        )
        #self.disabled_websocket_echo = self.create_gobj(
        #    'broadcast',
        #    GSockjsServer,
        #    self,
        #    sockjs_app_class=BroadcastConnection,
        #)
